#!/usr/bin/env python3

# Copyright (without_goal+curr_emb) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
r"""
This file hosts task-specific or trainer-specific environments for trainers.
All environments here should be a (direct or indirect ) subclass of Env class
in habitat. Customized environments should be registered using
``@baseline_registry.register_env(name="myEnv")` for reusability
"""

from typing import Optional, Type

import habitat
from habitat import Config, Dataset
from custom_habitat_baselines.common.baseline_registry import baseline_registry
import os

def get_env_class(env_name: str) -> Type[habitat.RLEnv]:
    r"""Return environment class based on name.

    Args:
        env_name: name of the environment.

    Returns:
        Type[habitat.RLEnv]: env class.
    """
    return baseline_registry.get_env(env_name)

MEM_DEBUG = False
@baseline_registry.register_env(name="NavRLEnv")
class NavRLEnv(habitat.RLEnv):
    def __init__(self, config: Config, dataset: Optional[Dataset] = None):
        self._rl_config = config.RL
        self._core_env_config = config.TASK_CONFIG
        self._reward_measure_name = self._rl_config.REWARD_MEASURE
        self._success_measure_name = self._rl_config.SUCCESS_MEASURE

        self._previous_measure = None
        self._previous_action = None
        self.pid = os.getpid()
        self.current_process = psutil.Process(self.pid)
        self.memKB = self.current_process.memory_info()[0] / 2. ** 20
        super().__init__(self._core_env_config, dataset)

    def reset(self):
        self._previous_action = None
        observations = super().reset()
        self._previous_measure = self._env.get_metrics()[
            self._reward_measure_name
        ]
        if MEM_DEBUG:
            memKB = self.current_process.memory_info()[0] / 2. ** 20
            scene = self.current_episode.scene_id.split('/')[-2]
            episode_id = self.current_episode.episode_id

            print("reset ENV %d %s %s stepp Current memory KB   : %9.3f KB"%(self.pid, scene, episode_id, memKB))
        return observations

    def step(self, *args, **kwargs):
        self._previous_action = kwargs["action"]
        if MEM_DEBUG:
            memKB = self.current_process.memory_info()[0] / 2. ** 20
            scene = self.current_episode.scene_id.split('/')[-2]
            episode_id = self.current_episode.episode_id
            if (memKB -self.memKB) > 10.0:
                print("====================ENV %d %s %s stepp Current memory KB   : %9.3f KB"%(self.pid, scene, episode_id, memKB))
            self.memKB = memKB
        return super().step(*args, **kwargs)

    def get_reward_range(self):
        return (
            self._rl_config.SLACK_REWARD - 1.0,
            self._rl_config.SUCCESS_REWARD + 1.0,
        )

    def get_reward(self, observations):
        reward = self._rl_config.SLACK_REWARD

        current_measure = self._env.get_metrics()[self._reward_measure_name]

        reward += self._previous_measure - current_measure
        self._previous_measure = current_measure

        if self._episode_success():
            reward += self._rl_config.SUCCESS_REWARD

        return reward

    def _episode_success(self):
        return self._env.get_metrics()[self._success_measure_name]

    def get_done(self, observations):
        done = False
        if self._env.episode_over or self._episode_success():
            done = True
        return done

    def get_info(self, observations):
        return self.habitat_env.get_metrics()
